@@ -1,15 +1,22 @@
 from copy import deepcopy
+import re
 import warnings
 from lxml.etree import Element
 from lxml import etree
 from zipfile import ZipFile, ZIP_DEFLATED
+from random import randint
 import shlex
+import mimetypes
+import os.path
 
 
 NAMESPACES = {
     'w': 'http://schemas.openxmlformats.org/wordprocessingml/2006/main',
     'mc': 'http://schemas.openxmlformats.org/markup-compatibility/2006',
     'ct': 'http://schemas.openxmlformats.org/package/2006/content-types',
+    'wp': 'http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing',
+    'a': 'http://schemas.openxmlformats.org/drawingml/2006/main',
+    'pic': 'http://schemas.openxmlformats.org/drawingml/2006/picture',
 }
 
 CONTENT_TYPES_PARTS = (
@@ -23,10 +30,17 @@ CONTENT_TYPE_SETTINGS = 'application/vnd.openxmlformats-officedocument.wordproce
 
 class MailMerge(object):
     def __init__(self, file, remove_empty_tables=False):
+        self.file = file
         self.zip = ZipFile(file)
         self.parts = {}
         self.settings = None
         self._settings_info = None
+
+        self.media = {}         # new images to add indexed by embed id,还有记录拓展名
+        self.rels = None        # etree for relationships
+        self._rels_info = None  # zi info block for rels
+        self.RELS_NAMESPACES = {'ns': None, 'od': None}
+
         self.remove_empty_tables = remove_empty_tables
 
         try:
@@ -38,8 +52,17 @@ class MailMerge(object):
                 elif type == CONTENT_TYPE_SETTINGS:
                     self._settings_info, self.settings = self.__get_tree_of_file(file)
 
+            # get the rels for image mappings
+            try:
+                self._rels_info, self.rels = self.__get_tree_of_file('word/_rels/document.xml.rels')
+                self.RELS_NAMESPACES['ns'] = self.rels.getroot().nsmap.get(None)
+                self.RELS_NAMESPACES['od'] = self.rels.getroot().nsmap.get(None).replace('package', 'officeDocument')
+            except:
+                pass
+
             to_delete = []
 
+            r = re.compile(r' MERGEFIELD +"?([^ ]+?)"? +(|\\\* MERGEFORMAT )', re.I)
             for part in self.parts.values():
 
                 for parent in part.findall('.//{%(w)s}fldSimple/..' % NAMESPACES):
@@ -48,10 +71,14 @@ class MailMerge(object):
                             continue
                         instr = child.attrib['{%(w)s}instr' % NAMESPACES]
 
-                        name = self.__parse_instr(instr)
-                        if name is None:
+                        # name = self.__parse_instr(instr)
+                        # if name is None:
+                        #     continue
+                        # parent[idx] = Element('MergeField', name=name)
+                        m = r.match(instr)
+                        if m is None:
                             continue
-                        parent[idx] = Element('MergeField', name=name)
+                        parent[idx] = Element('MergeField', name=m.group(1))
 
                 for parent in part.findall('.//{%(w)s}instrText/../..' % NAMESPACES):
                     children = list(parent)
@@ -80,11 +107,15 @@ class MailMerge(object):
                         for instr in instr_elements[1:]:
                             instr.getparent().remove(instr)
 
-                        name = self.__parse_instr(instr_text)
-                        if name is None:
-                            continue
+                        # name = self.__parse_instr(instr_text)
+                        # if name is None:
+                        #     continue
 
-                        parent[idx_begin] = Element('MergeField', name=name)
+                        # parent[idx_begin] = Element('MergeField', name=name)
+                        m = r.match(instr_text)
+                        if m is None:
+                            continue
+                        parent[idx_begin] = Element('MergeField', name=m.group(1))
 
                         # use this so we know *where* to put the replacement
                         instr_elements[0].tag = 'MergeText'
@@ -119,25 +150,76 @@ class MailMerge(object):
         return name
 
     def __get_tree_of_file(self, file):
-        fn = file.attrib['PartName' % NAMESPACES].split('/', 1)[1]
+        # fn = file.attrib['PartName' % NAMESPACES].split('/', 1)[1]
+        if isinstance(file, etree._Element):
+            fn = file.get('PartName').split('/', 1)[1]
+        else:
+            fn = file
         zi = self.zip.getinfo(fn)
         return zi, etree.parse(self.zip.open(zi))
 
     def write(self, file):
         # Replace all remaining merge fields with empty values
+        # Remove all remaining image with descr="delete"
         for field in self.get_merge_fields():
             self.merge(**{field: ''})
 
+        reservedIdList = []
+        parts = self.parts.values()
+        for part in parts:
+            for image in part.findall('.//wp:docPr/..', namespaces=NAMESPACES):
+                picNode = image.find('.//wp:docPr', namespaces=NAMESPACES)
+                if "descr" not in picNode.attrib.keys() or picNode.attrib['descr'] != 'deletable':
+                    embed_node = image.find('.//a:blip', namespaces=NAMESPACES)
+                    embed_attr = embed_node.attrib.keys()[0]
+                    imageId = embed_node.attrib[embed_attr]
+                    if "rId" in imageId:
+                        reservedIdList.append(imageId)
+        # Id get, all Ids.  then check if has not deletable image exists, then keep write the image to new file
+        regStr = r'<[^<]*Id="(rId\d+)"[^<]*Target="([^<]*)"/>'
+        r = re.compile(regStr)
+        with self.zip.open("word/_rels/document.xml.rels") as relFile:
+            content = relFile.read().decode("utf8")
+        res = r.findall(content)
+        # keep rId,target as dict
+        idTarget = dict(res)
+        targetList = [idTarget[rId] for rId in reservedIdList]
+
         with ZipFile(file, 'w', ZIP_DEFLATED) as output:
             for zi in self.zip.filelist:
+                zfname = zi.filename
                 if zi in self.parts:
                     xml = etree.tostring(self.parts[zi].getroot())
+                    # remove delete image's w:drawing xml block from document.xml
+                    if zfname == "word/document.xml":
+                        docContent = xml.decode('utf8')
+                        reg = re.compile(r'<w:drawing>(?=.*descr).+?<\/w:drawing>')
+                        # regex too hard…… so, use __reFun function add condition to remove
+                        xml = reg.sub(self.__reFun, docContent)
                     output.writestr(zi.filename, xml)
                 elif zi == self._settings_info:
                     xml = etree.tostring(self.settings.getroot())
                     output.writestr(zi.filename, xml)
+                elif zi == self._rels_info:
+                    xml = etree.tostring(self.rels.getroot())
+                    output.writestr(zi.filename, xml)
                 else:
-                    output.writestr(zi.filename, self.zip.read(zi))
+                    if zfname.endswith((".bmp", ".png", ".jpeg", ".jpg", ".gif", ".ico")) and zfname.split("word/")[-1] not in targetList:
+                        # not reserved image, no need to write
+                        pass
+                    else:
+                        output.writestr(zi.filename, self.zip.read(zi))
+            # add new images to media folder is we have images merged
+            for img_id_extension, img_data in self.media.items():
+                arr = img_id_extension.split(':')
+                output.writestr('media/{}.{}'.format(arr[0], arr[1]), img_data)
+
+    def __reFun(self, obj):
+        str = obj.group()
+        if 'deletable' in str:
+            return ''
+        else:
+            return str
 
     def get_merge_fields(self, parts=None):
         if not parts:
@@ -152,7 +234,7 @@ class MailMerge(object):
         """
         Duplicate template. Creates a copy of the template, does a merge, and separates them by a new paragraph, a new break or a new section break.
         separator must be :
-        - page_break : Page Break. 
+        - page_break : Page Break.
         - column_break : Column Break. ONLY HAVE EFFECT IF DOCUMENT HAVE COLUMNS
         - textWrapping_break : Line Break.
         - continuous_section : Continuous section break. Begins the section on the next paragraph.
@@ -265,6 +347,48 @@ class MailMerge(object):
                     self.__merge_field(part, field, replacement)
 
     def __merge_field(self, part, field, text):
+        if field.endswith('_source'):
+            if os.path.exists(text):
+                # 此处检查确认是图片。如果是，读取到text变量。如果不是，删除原有的placeholder图片(save的时候查找删除)
+                mimetype = mimetypes.guess_type(text)[0]
+                if mimetype in ['image/bmp', 'image/jpeg', 'image/png', 'image/gif', 'image/x-icon']:
+                    # 这个时候，text为文件绝对地址
+                    img_name = field
+                    # inline_img_el = part.find(
+                    #     './/wp:docPr[@title="{}"]/..'.format("IMAGE:" + img_name), namespaces=NAMESPACES)
+                    imglist = part.findall(
+                        './/wp:docPr[@title="{}"]/..'.format("IMAGE:" + img_name), namespaces=NAMESPACES)
+                    if len(imglist) > 0:
+                        for inline_img_el in imglist:
+                            # if inline_img_el:
+                            embed_node = inline_img_el.find('.//a:blip', namespaces=NAMESPACES)
+                            if embed_node is not None:
+                                # generate a random id and add tp media list for later export to media folder in zip file
+                                img_id = 'MMR{}'.format(randint(10000000, 999999999))
+                                with open(text, 'rb') as f:
+                                    content = f.read()
+                                extension = text.split('.')[-1]
+                                self.media[img_id + ':' + extension] = content
+                                # self.media[extension] = extension
+
+                                # add a relationship
+                                last_img_relationship = self.rels.findall(
+                                    '{%(ns)s}Relationship[@Type="%(od)s/image"]' % self.RELS_NAMESPACES)[-1]
+                                new_img_relationship = deepcopy(last_img_relationship)
+                                new_img_relationship.set('Id', img_id)
+                                new_img_relationship.set('Target', '/media/{}.{}'.format(img_id, extension))
+                                self.rels.getroot().append(new_img_relationship)
+
+                                # replace the embed attrib with the new image_id
+                                embed_node = inline_img_el.find('.//a:blip', namespaces=NAMESPACES)
+                                embed_attr = embed_node.attrib.keys()[0]
+                                embed_node.attrib[embed_attr] = img_id
+                            # mark as done
+                            inline_img_el.find(
+                                'wp:docPr', namespaces=NAMESPACES).attrib['title'] = 'replaced_image_{}'.format(img_id)
+                            inline_img_el.find('wp:docPr', namespaces=NAMESPACES).attrib['descr'] = ''
+            return
+
         for mf in part.findall('.//MergeField[@name="%s"]' % field):
             children = list(mf)
             mf.clear()  # clear away the attributes
@@ -327,14 +451,6 @@ class MailMerge(object):
                 return list
         return []
 
-    def check_finger_print(self, fingerPrint):
-        parts = self.parts.values()
-        for part in parts:
-            mf = part.xpath('//p[contain(text(),"{%s}")]' % fingerPrint)
-            if mf is not None:
-                return True
-        return False
-
     def __enter__(self):
         return self
 
